@@ -3,12 +3,18 @@ import logging
 from couchdbkit import ResourceNotFound
 from couchdbkit.ext.django import schema
 import datetime
+import sqlalchemy
 from dimagi.utils.parsing import json_format_date
 from dimagi.utils.read_only import ReadOnlyObject
 from fluff import exceptions
-from pillowtop.listener import BasicPillow
+from fluff.signals import BACKEND_SQL, BACKEND_COUCH
+from pillowtop.listener import PythonPillow
 from .signals import indicator_document_updated
-import fluff.sync_couchdb
+try:
+    import fluff.sync_couchdb
+except ImportError:
+    pass
+import fluff.util
 
 
 REDUCE_TYPES = set(['sum', 'count', 'min', 'max', 'sumsqr'])
@@ -161,6 +167,9 @@ class Calculator(object):
                     list(fn(item))
                     if passes_filter else []
                 )
+                for val in values[slug]:
+                    if val['group_by'] and len(val['group_by']) != len(self.fluff.group_by):
+                        raise Exception("group_by returned by emitter is of different length to default group_by")
             except Exception:
                 logging.exception((
                     "Error in emitter %s > %s > %s: '%s'"
@@ -173,17 +182,20 @@ class Calculator(object):
                 raise
         return values
 
-    def get_result(self, key, date_range=None, reduce=True):
+    def get_result(self, key, date_range=None, reduce=True, verbose_results=False):
         """
         If your Calculator does not have a window set, you must pass a tuple of
         date or datetime objects to date_range
         """
-        if self.window:
+        if verbose_results:
+            assert not reduce, "can't have reduce set for verbose results"
+
+        if date_range is not None:
+            start, end = date_range
+        elif self.window:
             now = self.fluff.get_now()
             start = now - self.window
             end = now
-        elif date_range is not None:
-            start, end = date_range
         result = {}
         for emitter_name in self._fluff_emitters:
             shared_key = [self.fluff._doc_type] + key + [self.slug, emitter_name]
@@ -221,18 +233,28 @@ class Calculator(object):
                 except IndexError:
                     result[emitter_name] = 0
             else:
+                # clean ids
                 def strip(id_string):
                     prefix = '%s-' % self.fluff.__name__
                     assert id_string.startswith(prefix)
                     return id_string[len(prefix):]
-                result[emitter_name] = [strip(row['id']) for row in q]
+                for row in q:
+                    row['id'] = strip(row['id'])
+
+                if not verbose_results:
+                    # strip down to ids
+                    result[emitter_name] = [row['id'] for row in q]
+                else:
+                    result[emitter_name] = q
+
         return result
 
-    def aggregate_results(self, keys, reduce=True):
+    def aggregate_results(self, keys, reduce=True, verbose_results=False, date_range=None):
 
         def iter_results():
             for key in keys:
-                result = self.get_result(key, reduce=reduce)
+                result = self.get_result(key, reduce=reduce, date_range=date_range,
+                                         verbose_results=verbose_results)
                 for slug, value in result.items():
                     yield slug, value
 
@@ -240,10 +262,14 @@ class Calculator(object):
             results = defaultdict(int)
             for slug, value in iter_results():
                 results[slug] += value
-        else:
+        elif not verbose_results:
             results = defaultdict(set)
             for slug, value in iter_results():
                 results[slug].update(value)
+        else:
+            results = defaultdict(list)
+            for slug, value in iter_results():
+                results[slug].extend(value)
 
         return results
 
@@ -267,7 +293,9 @@ class FlatField(schema.StringProperty):
         super(FlatField, self).__init__(*args, **kwargs)
 
     def calculate(self, item):
-        return self.fn(item)
+        result = self.fn(item)
+        assert isinstance(result, basestring)
+        return result
 
 
 class AttributeGetter(object):
@@ -304,7 +332,11 @@ class IndicatorDocumentMeta(schema.DocumentMeta):
             calculator.fluff = cls
             calculator.slug = slug
         cls._calculators = calculators
-        cls._flat_fields = flat_fields 
+        cls._flat_fields = flat_fields
+
+        instance = cls()
+        if instance.save_direct_to_sql:
+            cls._table = util.get_indicator_model(name, instance)
         return cls
 
 
@@ -317,6 +349,7 @@ class IndicatorDocument(schema.Document):
     wrapper = None
     document_filter = None
     group_by = ()
+    save_direct_to_sql = None
 
     # Mapping of group_by field to type. Used to communicate expected type in fluff diffs.
     # See ALL_TYPES
@@ -340,6 +373,17 @@ class IndicatorDocument(schema.Document):
     def get_group_values(self):
         return [self[attr] for attr in self.get_group_names()]
 
+    def get_group_types(self):
+        group_by_type_map = self.group_by_type_map or {}
+        for gb in self.wrapped_group_by:
+            attrib = gb.attribute
+            if attrib not in group_by_type_map:
+                group_by_type_map[attrib] = TYPE_STRING
+            else:
+                assert group_by_type_map[attrib] in ALL_TYPES
+
+        return group_by_type_map
+
     @classmethod
     def get_now(cls):
         return datetime.datetime.utcnow().date()
@@ -418,20 +462,12 @@ class IndicatorDocument(schema.Document):
         if not diff_keys:
             return None
 
-        group_by_type_map = self.group_by_type_map or {}
-        for gb in self.wrapped_group_by:
-            attrib = gb.attribute
-            if attrib not in group_by_type_map:
-                group_by_type_map[attrib] = TYPE_STRING
-            else:
-                assert group_by_type_map[attrib] in ALL_TYPES
-
         diff = dict(domains=list(self.domains),
                     database=self.Meta.app_label,
                     doc_type=self._doc_type,
                     group_names=self.get_group_names(),
                     group_values=self.get_group_values(),
-                    group_type_map=group_by_type_map,
+                    group_type_map=self.get_group_types(),
                     indicator_changes=[],
                     all_indicators=[])
         indicator_changes = diff["indicator_changes"]
@@ -517,6 +553,66 @@ class IndicatorDocument(schema.Document):
         changed = set(o for o in intersect if left[o] != right[o])
         return added | removed | changed
 
+    def save_to_sql(self, diff, engine):
+        if not diff:
+            # empty indicator document
+            return
+
+        default_key = (self.id,) + tuple(diff['group_values'])
+        rows = {}
+
+        def set_row_val(rowkey, col_name, col_value):
+            row = rows.setdefault(rowkey, {})
+            row[col_name] = col_value
+
+        flat_keys = None
+        try:
+            flat_keys = self._flat_fields.keys()
+        except AttributeError:
+            pass
+
+        for change in diff['indicator_changes']:
+            name = '{0}_{1}'.format(change['calculator'], change['emitter'])
+            for value_dict in change['values']:
+                value = value_dict['value']
+                group_by = value_dict['group_by']
+                date = value_dict['date']
+                if group_by:
+                    key = (self.id,) + tuple(group_by) + (date,)
+                else:
+                    key = default_key + (date,)
+                set_row_val(key, name, value)
+                for flat_key in flat_keys:
+                    set_row_val(key, flat_key, self[flat_key])
+
+        types = self.get_group_types()
+        types['date'] = 'date'
+        names = ['doc_id'] + self.get_group_names() + ['date']
+        connection = engine.connect()
+        try:
+            # delete all existing rows for this doc to ensure we aren't left with stale data
+            delete = self._table.delete(self._table.c.doc_id == self.id)
+            connection.execute(delete)
+
+            for key, columns in rows.items():
+                key_columns = dict(zip(names, key))
+                for name, value in key_columns.items():
+                    if value is None:
+                        key_columns[name] = util.default_null_value_placeholder(types[name])
+                all_columns = dict(key_columns.items() + columns.items())
+
+                try:
+                    insert = self._table.insert().values(**all_columns)
+                    connection.execute(insert)
+                except sqlalchemy.exc.IntegrityError:
+                    if columns:
+                        update = self._table.update().values(**columns)
+                        for k, v in key_columns.items():
+                            update = update.where(self._table.c[k] == v)
+                        connection.execute(update)
+        finally:
+            connection.close()
+
     @classmethod
     def pillow(cls):
         wrapper = cls.wrapper or cls.document_class
@@ -526,14 +622,15 @@ class IndicatorDocument(schema.Document):
             domains = ' '.join(cls.domains)
             extra_args['domains'] = domains
 
-        document_filter = cls.document_filter
         return type(FluffPillow)(cls.__name__ + 'Pillow', (FluffPillow,), {
-            'couch_filter': 'fluff_filter/domain_type',
             'extra_args': extra_args,
             'document_class': cls.document_class,
             'wrapper': wrapper,
             'indicator_class': cls,
-            'document_filter': document_filter,
+            'document_filter': cls.document_filter,
+            'domains': cls.domains,
+            'doc_type': doc_type,
+            'save_direct_to_sql': cls().save_direct_to_sql,
         })
 
     @classmethod
@@ -550,18 +647,43 @@ class IndicatorDocument(schema.Document):
         return calculator.get_result(key, date_range=date_range, reduce=reduce)
 
     @classmethod
-    def aggregate_results(cls, calc_name, keys, reduce=True):
+    def aggregate_results(cls, calc_name, keys, reduce=True, date_range=None):
         calculator = cls.get_calculator(calc_name)
-        return calculator.aggregate_results(keys, reduce=reduce)
+        return calculator.aggregate_results(keys, reduce=reduce, date_range=date_range)
+
+    @classmethod
+    def aggregate_all_results(cls, keys, reduce=True, date_range=None):
+        return dict(
+            (calc_name, calc.aggregate_results(keys, reduce=reduce, date_range=date_range))
+            for calc_name, calc in cls._calculators.items()
+        )
 
     class Meta:
         app_label = 'fluff'
 
-
-class FluffPillow(BasicPillow):
+class FluffPillow(PythonPillow):
     document_filter = None
     wrapper = None
     indicator_class = IndicatorDocument
+    domains = None
+    doc_type = None
+    save_direct_to_sql = False
+
+    @classmethod
+    def get_sql_engine(cls):
+        engine = getattr(cls, '_engine', None)
+        if not engine:
+            import sqlalchemy
+            from django.conf import settings
+            engine = sqlalchemy.create_engine(settings.SQL_REPORTING_DATABASE_URL)
+            cls._engine = engine
+        return engine
+
+    def python_filter(self, doc):
+        assert self.domains
+        assert None not in self.domains
+        assert self.doc_type is not None
+        return doc.get('domain', None) in self.domains and doc.get('doc_type', None) == self.doc_type
 
     def change_transform(self, doc_dict):
         doc = self.wrapper.wrap(doc_dict)
@@ -588,8 +710,12 @@ class FluffPillow(BasicPillow):
 
     def change_transport(self, indicators):
         old_indicator, new_indicator = indicators
-        new_indicator.save()
 
-        diff = new_indicator.diff(old_indicator)
-        if diff:
-            indicator_document_updated.send(sender=self, diff=diff)
+        diff = new_indicator.diff(None)  # pass in None for old_doc to force diff with ALL indicators
+        if self.save_direct_to_sql:
+            new_indicator.save_to_sql(diff, self.get_sql_engine())
+        else:
+            new_indicator.save()
+
+        backend = BACKEND_SQL if self.save_direct_to_sql else BACKEND_COUCH
+        indicator_document_updated.send(sender=self, doc_id=new_indicator.id, diff=diff, backend=backend)
